#pragma once
#include "elf.h"
#include "CDyn.h"
#include "CHeader.h"
#include "CPhdr.h"
#include "CShdr.h"
#include <vector>
#include <ptr_vector.h>

#define GET_ALIGNED_SIZE(requested_size, block_size) ((requested_size) % (block_size) == 0 ? (requested_size) : (((requested_size) / (block_size))+1)*(block_size))

namespace ELF_format {

	typedef void(*tProcessFileOffset)(uint64_t* pAddress, uint64_t* pSize, void* pUserData);

	class CELFmanager
	{
		eValueEncoding valueEncoding;

	public:
		CHeader header;
		stdx::ptr_vector<CDyn> dynamic_table;
		stdx::ptr_vector<CPhdr> program_header_table;
		stdx::ptr_vector<CShdr> section_header_table;
		char* pStringSectionData;

	public:
		// Constructor
		CELFmanager();

		// Deconstructor
		~CELFmanager();

		// Loads a file
		bool Load(const char* filename);

		// Saves a new file
		bool Save(const char* filename);

		// Updates file offsets
		void UpdateFileOffsets();

		// Returns value encoding
		eValueEncoding GetValueEncoding() { return this->valueEncoding; }

		// Returns offset to end of file
		uint64_t GetOffsetToEndOfFile();

		// Reads a part of memory from segment
		bool ReadPartOfMemoryFromSegment(uint64_t address, uint64_t size, void* pMemory);

		// Returns a pointer to memory from segment
		void* GetPointerToPartOfMemoryFromSegment(uint64_t address, uint64_t size);

		// Returns offset to end of memory
		uint64_t GetOffsetToEndOfMemory();

		// Returns the size of phdr entry
		uint32_t GetSizeOfShdr();

		// Returns the size of dyn entry
		uint32_t GetSizeOfDyn();

		// Returns iterator to section by address
		stdx::ptr_vector<CShdr>::iterator FindSectionByAddress(uint32_t address);

		// Returns iterator to first section with specified type 
		stdx::ptr_vector<CShdr>::iterator FindFirstSectionWithType(uint32_t sh_type);

		// Returns iterator to first section with specified type 
		stdx::ptr_vector<CShdr>::iterator FindFirstSectionWithName(const char* name);

		// Returns iterator to first program segment with specified type 
		stdx::ptr_vector<CPhdr>::iterator FindFirstProgramSegmentWithType(uint32_t p_type);

		// Returns iterator to first dynamic table entry with specified tpye
		stdx::ptr_vector<CDyn>::iterator FindFirstEntryInDynamicTableWithTag(int64_t d_tag);


	private:
		// Reads ELF headers
		template<class phdrType, class dynType> bool ReadProgramSegmentHeaders(FILE* fp);

		// Writes ELF program segments
		template<class phdrType, class dynType> bool WriteProgramSegmentHeaders(FILE* fp);

		// Reads section headers headers
		template<class T> bool ReadSectionHeaders(FILE* fp);

		// Writes section headers headers
		template<class T> bool WriteSectionHeaders(FILE* fp);

		// Processes file offsets
		void ProcessFileOffsets(tProcessFileOffset func, void* pUserData);

	public:
		// Removes file bytes from segment
		void RemoveFileBytesFromSegment(uint64_t address, uint64_t size);
	};
}